/*
  ESP_nacitani_pulsu_elektromeru_wifimanager.ino
  Wifimanager + nacitani pulsu 3x elektromeru a odesilani dat na EmonCms server
  v config menu nastaveni nodeID, imp/kWh a pocatecni stav ve Wh pro 3 elmery
  ukladani do FRAM pamÄ›ti FM16CL16B
  Spusteni konfigurace tlacitkem na D3 (FLASH)
  Format SPIFFS pameti, vynulovani registru impulsu a Wh, reset WiFi nastaveni
  jumperem na A0
  Pulsy HIGH
  Arduino IDE: 1.8.5
  podpora ESP8266 v 2.4.1   https://github.com/esp8266/Arduino
  NodeMCU v1.0 (WeMos mini Pro),80MHz,4M(1M SPIFFS)
  WeMos mini Lite,80MHz,1M(128K SPIFFS)

  6.2.2022
  Tomas Nevrela - http://tnweb.tode.cz/
*/

#include <FS.h>
#include <ESP8266WiFi.h>          //https://github.com/esp8266/Arduino
#include <ArduinoJson.h>          //https://github.com/bblanchon/ArduinoJson
#include <ESP8266WebServer.h>
#include <DNSServer.h>
#include <WiFiManager.h>          //https://github.com/kentaylor/WiFiManager
#include <FRAMAnything.h>// knihovna pro FRAM FM24c16


#define D1 5 //
#define D2 4 // 
#define D3 0 //pripojeno tlacitko FLASH, spusteni konfigurace wifi
#define D4 2 //LED
#define D5 14 //pulsy 1
#define D6 12 //pulsy 2
#define D7 13 //pulsy 3
#define D8 15
#define pulsePin_1 D5 // pripojene pulsy elektromeru 1
#define pulsePin_2 D6 // pripojene pulsy elektromeru 2
#define pulsePin_3 D7 // pripojene pulsy elektromeru 3
#define RESET_PIN A0 //  GND -> format SPIFFS, reset WiFi nastaveni

const byte DEVADDR = 0x50;// I2C adresa FM24C16
const int PIN_LED = 2; // D4 ,LOW -> LED Sviti.
const int TRIGGER_PIN = 0; // D3 LOW -> spusteni configurace WiFi

const char* CONFIG_FILE = "/config.json";

// Indicates whether ESP has WiFi credentials saved from previous session
bool initialConfig = false;

// Default configuration values
char emoncmsAddress[40] = "emoncms.org";// 192.168.xx.xx
word port = 80;
char emoncmsApiKey[35] = "xxxxxxxxxx";

volatile uint32_t pulseCount1 = 0, pulseCount2 = 0, pulseCount3 = 0 ;
volatile uint32_t oldPulseCount1 = 0, oldPulseCount2 = 0, oldPulseCount3 = 0 ;
volatile unsigned int lastBlinkmillis1 = 0, lastBlinkmillis2 = 0, lastBlinkmillis3 = 0 ;

uint16_t perKomunikace = 10; // perioda odesilani dat na emoncms server (s)
uint32_t currentMillis = 0;
uint32_t lastMillis, SerlastMillis;
bool readConfigFile();
bool writeConfigFile();
uint16_t pulse_factor1 = 1000;
uint16_t pulse_factor2 = 1000;
uint16_t pulse_factor3 = 1000;
double fWh1, fWh2, fWh3;
double oldfWh1, oldfWh2, oldfWh3;
uint32_t Wh1, Wh2, Wh3;
float Whpp1, Whpp2, Whpp3;
uint8_t node_id = 5; // default ID nodu (Elektromery_5)

//*******************************************************************************************
void ICACHE_RAM_ATTR pulseCounter_1()
{
  unsigned int newBlinkmillis = millis();
  unsigned int intervalmillis = newBlinkmillis - lastBlinkmillis1;
  delayMicroseconds(10);
  if (intervalmillis > 10) {
    if (digitalRead(pulsePin_1) == HIGH) {
      pulseCount1++;
    }
  }
  lastBlinkmillis1 = newBlinkmillis;
}
//*******************************************************************************************
void ICACHE_RAM_ATTR pulseCounter_2()
{
  unsigned int newBlinkmillis = millis();
  unsigned int intervalmillis = newBlinkmillis - lastBlinkmillis2;
  delayMicroseconds(10);
  if (intervalmillis > 10) {
    if (digitalRead(pulsePin_2) == HIGH) {
      pulseCount2++;
    }
  }
  lastBlinkmillis2 = newBlinkmillis;
}
//*****************************************************************************************
void ICACHE_RAM_ATTR pulseCounter_3()
{
  unsigned int newBlinkmillis = millis();
  unsigned int intervalmillis = newBlinkmillis - lastBlinkmillis3;

  delayMicroseconds(10);
  if (intervalmillis > 10) {
    if (digitalRead(pulsePin_3) == HIGH) {
      pulseCount3++;
    }
  }
  lastBlinkmillis3 = newBlinkmillis;
}
//==================================SETUP================================================
void setup() {

  Serial.begin(115200);
  Wire.begin();
  Serial.println("\n Starting");

  pinMode(PIN_LED, OUTPUT);
  // Initialize trigger pin
  pinMode(TRIGGER_PIN, INPUT_PULLUP);


  pinMode(pulsePin_1, INPUT);
  pinMode(pulsePin_2, INPUT);
  pinMode(pulsePin_3, INPUT);

  attachInterrupt(digitalPinToInterrupt(D5), pulseCounter_1, RISING);
  attachInterrupt(digitalPinToInterrupt(D6), pulseCounter_2, RISING);
  attachInterrupt(digitalPinToInterrupt(D7), pulseCounter_3, RISING);


  // Mount the filesystem
  bool result = SPIFFS.begin();
  Serial.println("SPIFFS opened: " + result);

  if (!readConfigFile()) {
    Serial.println("Failed to read configuration file, using default values");
  }

  //FRAM_WriteAnything(DEVADDR, 4, 99999900);//TEST

  // FRAM_WriteAnything(DEVADDR, 0, 0);// TEST
  // FRAM_WriteAnything(DEVADDR, 4, 0);
  // FRAM_WriteAnything(DEVADDR, 12, 0);
  // FRAM_WriteAnything(DEVADDR, 16, 0);
  // FRAM_WriteAnything(DEVADDR, 24, 0);
  // FRAM_WriteAnything(DEVADDR, 28, 0);

  // for (short i = 0; i < 2048; i++) {  //TEST smaze celou FRAM
  // FRAM_WriteAnything(DEVADDR, i, 0);
  // }


  FRAM_ReadAnything(DEVADDR, 0, pulseCount1);//
  FRAM_ReadAnything(DEVADDR, 4, fWh1);
  oldPulseCount1 = pulseCount1;
  FRAM_ReadAnything(DEVADDR, 12, pulseCount2);
  FRAM_ReadAnything(DEVADDR, 16, fWh2);
  oldPulseCount2 = pulseCount2;
  FRAM_ReadAnything(DEVADDR, 24, pulseCount3);
  FRAM_ReadAnything(DEVADDR, 28, fWh3);
  oldPulseCount3 = pulseCount3;

  Serial.print(" ESP ID Chip: ");
  Serial.println (ESP.getChipId());


  WiFi.printDiag(Serial); //Remove this line if you do not want to see WiFi password printed
  if (WiFi.SSID() == "") {
    Serial.println("Zadne prihlasovaci udaje k pristupovemu bodu");
    initialConfig = true;
  }
  else {
    digitalWrite(PIN_LED, HIGH); // Turn led off as we are not in configuration mode.
    WiFi.mode(WIFI_STA); // Force to station mode because if device was switched off while in access point mode it will start up next time in access point mode.
    unsigned long startedAt = millis();
    Serial.print("After waiting ");
    int connRes = WiFi.waitForConnectResult();
    float waited = (millis() - startedAt);
    Serial.print(waited / 1000);
    Serial.print(" secs in setup() connection result is ");
    Serial.println(connRes);
  }

  if (WiFi.status() != WL_CONNECTED) {
    Serial.println("failed to connect, finishing setup anyway");
  } else {
    Serial.print("local ip: ");
    Serial.println(WiFi.localIP());
  }
}

//=======================================LOOP=============================================
void loop() {
  // is configuration portal requested?


  if ((digitalRead(TRIGGER_PIN) == LOW) || (initialConfig)) {
    Serial.println("Configuration portal requested.");
    digitalWrite(PIN_LED, LOW); // turn the LED on by making the voltage LOW to tell us we are in configuration mode.
    //Local intialization. Once its business is done, there is no need to keep it around

    // Extra parameters to be configured
    // After connecting, parameter.getValue() will get you the configured value
    // Format: <ID> <Placeholder text> <default value> <length> <custom HTML> <label placement>
    WiFiManagerParameter p_emoncmsAddress("emoncmsaddress", "EmonCMS Adresa", emoncmsAddress, 40);

    char convertedValue[9];
    sprintf(convertedValue, "%d", port);
    WiFiManagerParameter p_port("port", "Port", convertedValue, 4);
    // Emoncms API Key - this is a straight forward string parameter
    WiFiManagerParameter p_emoncmsApiKey("emoncmsapikey", "Emoncms-API Key", emoncmsApiKey, 33);

    sprintf(convertedValue, "%d", node_id);
    WiFiManagerParameter p_node_id("node_id", "ID", convertedValue, 2);

    sprintf(convertedValue, "%d", pulse_factor1);
    WiFiManagerParameter p_pulseFact1("pulseFactor1", "E1 imp/kWh", convertedValue, 5);
    sprintf(convertedValue, "%d", pulse_factor2);
    WiFiManagerParameter p_pulseFact2("pulseFactor2", "E2 imp/kWh", convertedValue, 5);
    sprintf(convertedValue, "%d", pulse_factor3);
    WiFiManagerParameter p_pulseFact3("pulseFactor3", "E3 imp/kWh", convertedValue, 5);
    sprintf(convertedValue, "%d", Wh1);
    WiFiManagerParameter p_Wh1("Wh1", "E1 Wh", convertedValue, 9);
    sprintf(convertedValue, "%d", Wh2);
    WiFiManagerParameter p_Wh2("Wh2", "E2 Wh", convertedValue, 9);
    sprintf(convertedValue, "%d", Wh3);
    WiFiManagerParameter p_Wh3("Wh3", "E3 Wh", convertedValue, 9);

    // Just a quick hint
    WiFiManagerParameter p_hint("<small>*Hint: if you want to reuse the currently active WiFi credentials, leave SSID and Password fields empty</small>");
    // Initialize WiFIManager
    WiFiManager wifiManager;

    if ((analogRead(RESET_PIN) <= 100)) { // pin reset nastaveni a WiFi
      SPIFFS.format(); //
      Serial.println("Reset PIN aktivovan - Format SPIFFS a reset nastaveni WiFi");
      for (short i = 0; i < 2048; i++) {  //smaze celou FRAM
        FRAM_WriteAnything(DEVADDR, i, 0);
      }
      wifiManager.resetSettings();
    }

    //add all parameters here

    wifiManager.addParameter(&p_hint);
    wifiManager.addParameter(&p_emoncmsAddress);
    wifiManager.addParameter(&p_port);
    wifiManager.addParameter(&p_emoncmsApiKey);
    wifiManager.addParameter(&p_node_id);
    wifiManager.addParameter(&p_pulseFact1);
    wifiManager.addParameter(&p_pulseFact2);
    wifiManager.addParameter(&p_pulseFact3);
    wifiManager.addParameter(&p_Wh1);
    wifiManager.addParameter(&p_Wh2);
    wifiManager.addParameter(&p_Wh3);

    //sets timeout in seconds until configuration portal gets turned off.
    //If not specified device will remain in configuration mode until
    //switched off via webserver or device is restarted.
    //wifiManager.setConfigPortalTimeout(600);

    // It starts an access point
    // and goes into a blocking loop awaiting configuration.
    // Once the user leaves the portal with the exit button
    // processing will continue
    if (!wifiManager.startConfigPortal("ESP_ELM")) {
      Serial.println("Nepripojeno k WiFi ale presto pokracujeme.");
    } else {
      // If you get here you have connected to the WiFi
      Serial.println("Pripojeno... :)");
    }

    // Getting posted form values and overriding local variables parameters
    // Config file is written regardless the connection state
    strcpy(emoncmsAddress, p_emoncmsAddress.getValue());
    port = atoi(p_port.getValue());
    strcpy(emoncmsApiKey, p_emoncmsApiKey.getValue());
    node_id = atoi(p_node_id.getValue());
    pulse_factor1 = atoi(p_pulseFact1.getValue());
    pulse_factor2 = atoi(p_pulseFact2.getValue());
    pulse_factor3 = atoi(p_pulseFact3.getValue());
    Wh1 = atoi(p_Wh1.getValue());
    Wh2 = atoi(p_Wh2.getValue());
    Wh3 = atoi(p_Wh3.getValue());

    fWh1 = Wh1;
    fWh2 = Wh2;
    fWh3 = Wh3;
    FRAM_WriteAnything(DEVADDR, 4, fWh1);
    FRAM_WriteAnything(DEVADDR, 16, fWh2);
    FRAM_WriteAnything(DEVADDR, 28, fWh3);

    writeConfigFile();

    digitalWrite(PIN_LED, HIGH); // Turn led off as we are not in configuration mode.

    ESP.reset(); // This is a bit crude. For some unknown reason webserver can only be started once per boot up
    // so resetting the device allows to go back into config mode again when it reboots.
    delay(5000);

  }
  //===================================================================================

  if (pulseCount1 != oldPulseCount1) {
    unsigned int incPulse = (pulseCount1 - oldPulseCount1);
    float IncWh = ((float)incPulse * Whpp1);
    fWh1 += IncWh;
    if (fWh1 >= 100000000.0) fWh1 = 0;
    oldPulseCount1 = pulseCount1;
    oldfWh1 = fWh1;
    Wh1 = fWh1;
    FRAM_WriteAnything(DEVADDR, 0, pulseCount1);
    FRAM_WriteAnything(DEVADDR, 4, fWh1);

  }


  //*****************************************************************

  if (pulseCount2 != oldPulseCount2) {
    unsigned int incPulse = (pulseCount2 - oldPulseCount2);
    float IncWh = ((float)incPulse * Whpp2);
    fWh2 += IncWh;
    if (fWh2 >= 100000000.0) fWh2 = 0;
    oldPulseCount2 = pulseCount2;
    oldfWh2 = fWh2;
    Wh2 = fWh2;
    FRAM_WriteAnything(DEVADDR, 12, pulseCount2);
    FRAM_WriteAnything(DEVADDR, 16, fWh2);
  }

  //***************************************************************



  if (pulseCount3 != oldPulseCount3) {
    unsigned int incPulse = (pulseCount3 - oldPulseCount3);
    float IncWh = ((float)incPulse * Whpp3);
    fWh3 += IncWh;
    if (fWh3 >= 100000000.0) fWh3 = 0;
    oldPulseCount3 = pulseCount3;
    oldfWh3 = fWh3;
    Wh3 = fWh3;
    FRAM_WriteAnything(DEVADDR, 24, pulseCount3);
    FRAM_WriteAnything(DEVADDR, 28, fWh3);
  }

  //***************************************************************
  currentMillis = millis();
  if ((unsigned long)(currentMillis - SerlastMillis >= 3000))
  {
    Serial.println();
    Serial.print(" ELM_1:");
    Serial.print(pulseCount1);
    Serial.print("imp ");
    Serial.print(fWh1, 2);
    Serial.print("Wh ");
    Serial.print(pulse_factor1);
    Serial.println("imp/kWh");

    Serial.print(" ELM_2:");
    Serial.print(pulseCount2);
    Serial.print("imp ");
    Serial.print(fWh2, 2);
    Serial.print("Wh ");
    Serial.print(pulse_factor2);
    Serial.println("imp/kWh");

    Serial.print(" ELM_3:");
    Serial.print(pulseCount3);
    Serial.print("imp ");
    Serial.print(fWh3, 2);
    Serial.print("Wh ");
    Serial.print(pulse_factor3);
    Serial.println("imp/kWh");
    Serial.println();

    SerlastMillis = currentMillis;
  }



  //=======================================================================================
  // Odeslani dat na server Emoncms //
  currentMillis = millis();
  if ((unsigned long)(currentMillis - lastMillis >= (perKomunikace * 1000)))
  {
    Serial.print("Connecting to ");
    Serial.println(emoncmsAddress);
    digitalWrite(PIN_LED, LOW);

    // Use WiFiClient class to create TCP connections
    WiFiClient client;

    if (!client.connect(emoncmsAddress, port)) {
      Serial.println("connection failed");
      return;
    }


    //int value= random(0,1000);
    //pulseCount = random(0,2000);

    long rssi = WiFi.RSSI();
    //*************************************************
    // prevod RSSI na kvalitu signalu 0-100%
    int quality = 0;
    if (rssi <= -100) {
      quality = 0;
    } else if (rssi >= -50) {
      quality = 100;
    } else {
      quality = 2 * (rssi + 100);
    }
    //*************************************************
    Wh1 = fWh1;
    Wh2 = fWh2;
    Wh3 = fWh3;



    //Build the JSON for emoncms to send data
    String json = "{Wifi:";
    json += quality;
    json += ",Elm1_pulse:";
    json += pulseCount1;
    json += ",Elm1_Wh:";
    json += Wh1;
    json += ",Elm2_pulse:";
    json += pulseCount2;
    json += ",Elm2_Wh:";
    json += Wh2;
    json += ",Elm3_pulse:";
    json += pulseCount3;
    json += ",Elm3_Wh:";
    json += Wh3;
    json += "}";

    //Build emoncms URL for sending data to
    String url = "/emoncms/input/post.json?node=";
    url += "Elektromery_"; //nodedata
    url += node_id;
    url += "&json=";
    url += json;
    url += "&apikey=";
    url += emoncmsApiKey;
    Serial.print(String("GET ") + url + " HTTP/1.1\r\n" +
                 "Host: " + emoncmsAddress + "\r\n" +
                 "Connection: close\r\n\r\n");

    client.print(String("GET ") + url + " HTTP/1.1\r\n" +
                 "Host: " + emoncmsAddress + "\r\n" +
                 "Connection: close\r\n\r\n");

    delay(10);
    digitalWrite(PIN_LED, HIGH);

    ///////////////// timeout //////////////////////
    unsigned long timeout = millis();
    while (client.available() == 0) {
      if (millis() - timeout > 5000) {
        Serial.println(">>> Client Timeout !");
        client.stop();
        return;
      }
    }

    // Read all the lines of the reply from server and print them to Serial
    while (client.available()) {
      String line = client.readStringUntil('\r');
      Serial.print(line);
    }

    Serial.println();
    Serial.println("closing connection");
    lastMillis = currentMillis;
  }
}

//=============================================================================
bool readConfigFile() {
  // this opens the config file in read-mode
  File f = SPIFFS.open(CONFIG_FILE, "r");

  if (!f) {
    Serial.println("Configuration file not found");
    return false;
  } else {
    // we could open the file
    size_t size = f.size();
    // Allocate a buffer to store contents of the file.
    std::unique_ptr<char[]> buf(new char[size]);

    // Read and store file contents in buf
    f.readBytes(buf.get(), size);
    // Closing file
    f.close();
    // Using dynamic JSON buffer which is not the recommended memory model, but anyway
    // See https://github.com/bblanchon/ArduinoJson/wiki/Memory%20model
    DynamicJsonDocument json(1024); //upraveno na json v6
    // Parse JSON string
    auto deserializeError = deserializeJson(json, buf.get());
    serializeJson(json, Serial);
    // Test if parsing succeeds.
    if (deserializeError) {
      Serial.println("JSON deserialize failed");
      return false;
    }
    //json.printTo(Serial);

    // Parse all config file parameters, override
    // local config variables with parsed values
    if (json.containsKey("emoncmsAddress")) {
      strcpy(emoncmsAddress, json["emoncmsAddress"]);
    }
    if (json.containsKey("emoncmsApiKey")) {
      strcpy(emoncmsApiKey, json["emoncmsApiKey"]);
    }
    if (json.containsKey("port")) {
      port = json["port"];
    }
    if (json.containsKey("node_id")) {
      node_id = json["node_id"];
    }
    if (json.containsKey("pulseFact1")) {
      pulse_factor1 = json["pulseFact1"];
      Whpp1 = ((float)1000 / (float)pulse_factor1);
    }
    if (json.containsKey("pulseFact2")) {
      pulse_factor2 = json["pulseFact2"];
      Whpp2 = ((float)1000 / (float)pulse_factor2);
    }
    if (json.containsKey("pulseFact3")) {
      pulse_factor3 = json["pulseFact3"];
      Whpp3 = ((float)1000 / (float)pulse_factor3);
    }
  }
  Serial.println("\nConfig file was successfully parsed");
  return true;
}
//=============================================================================
bool writeConfigFile() {
  Serial.println("Saving config file");
  DynamicJsonDocument json(1024);//upraveno na json v6


  // JSONify local configuration parameters
  json["emoncmsAddress"] = emoncmsAddress;
  json["port"] = port;
  json["node_id"] = node_id;
  json["emoncmsApiKey"] = emoncmsApiKey;
  json["pulseFact1"] = pulse_factor1;
  json["pulseFact2"] = pulse_factor2;
  json["pulseFact3"] = pulse_factor3;

  // Open file for writing
  File f = SPIFFS.open(CONFIG_FILE, "w");
  if (!f) {
    Serial.println("Failed to open config file for writing");
    return false;
  }

  serializeJson(json, Serial);
  serializeJson(json, f);
  f.close();

  Serial.println("\nConfig file was successfully saved");
  return true;
}



